package com.myoa.engineering.crawl.shopping.event.handler;

import com.myoa.engineering.crawl.shopping.domain.model.UserNotifyModel;
import com.myoa.engineering.crawl.shopping.domain.model.v2.AppUserModel;
import com.myoa.engineering.crawl.shopping.domain.model.v2.ArticleModel;
import com.myoa.engineering.crawl.shopping.domain.model.v2.SubscribedKeywordAggregatedModel;
import com.myoa.engineering.crawl.shopping.event.ArticleUpsertEvent;
import com.myoa.engineering.crawl.shopping.service.AppUserQueryService;
import com.myoa.engineering.crawl.shopping.service.SubscribedKeywordCacheService;
import com.myoa.engineering.crawl.shopping.service.slack.UserNotifyService;
import com.myoa.engineering.crawl.shopping.support.dto.constant.CrawlTarget;
import com.slack.api.methods.request.chat.ChatPostMessageRequest;
import com.slack.api.methods.response.chat.ChatPostMessageResponse;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

@Component
public class ArticleUpsertEventListener {

    private final SubscribedKeywordCacheService subscribedKeywordCacheService;
    private final AppUserQueryService appUserQueryService;
    private final UserNotifyService userNotifyService;


    public ArticleUpsertEventListener(SubscribedKeywordCacheService subscribedKeywordCacheService,
                                      AppUserQueryService appUserQueryService, UserNotifyService userNotifyService) {
        this.subscribedKeywordCacheService = subscribedKeywordCacheService;
        this.appUserQueryService = appUserQueryService;
        this.userNotifyService = userNotifyService;
    }


    @EventListener
    public void handleArticleUpsertEvent(ArticleUpsertEvent event) {
        Map<CrawlTarget, List<ArticleModel>> articleMap =
                ((List<ArticleModel>) event.getSource()).stream()
                                                        .collect(Collectors.groupingBy(ArticleModel::getCrawlTarget));
        Map<CrawlTarget, ChatPostMessageResponse> allArticleNotifiedResultMap =
                articleMap.entrySet()
                          .stream()
                          .collect(Collectors.toMap(Map.Entry::getKey, e -> notifyMessage(e.getKey(), e.getValue())));

        List<AppUserModel> appUsers = appUserQueryService.findAll();

        appUsers.stream()
                .filter(AppUserModel::getEnabled)
                .flatMap(user -> {
                    Map<CrawlTarget, SubscribedKeywordAggregatedModel> subscribedKeywords =
                            subscribedKeywordCacheService.getSubscribedKeywordsCached(user.getSlackId());
                    return subscribedKeywords.entrySet()
                                             .stream()
                                             .map(entry -> {
                                                 List<ArticleModel> filtered = doAhoCorasick(articleMap.get(entry.getKey())).apply(entry.getValue());
                                                 return UserNotifyModel.builder()
                                                                       .slackId(user.getSlackId())
                                                                       .articles(filtered)
                                                                       .chatPostMessageResponse(allArticleNotifiedResultMap.get(entry.getKey()))
                                                                       .build();
                                             });
                })
                .forEach(this::notifyMessage);
    }

    private Function<SubscribedKeywordAggregatedModel, List<ArticleModel>> doAhoCorasick(
            List<ArticleModel> articles) {
        return userTrieModel -> filterAhocorasick(articles, userTrieModel);
    }

    private List<ArticleModel> filterAhocorasick(List<ArticleModel> articles,
                                                 SubscribedKeywordAggregatedModel trieModel) {
        return articles.stream()
                       .filter(article -> !trieModel.getAhoCorasickTrie()
                                                    .parseText(article.getTitle())
                                                    .isEmpty())
                       .toList();
    }

    private ChatPostMessageResponse notifyMessage(CrawlTarget crawlTarget, List<ArticleModel> articles) {
        var sb = new StringBuilder();
        sb.append("[").append(crawlTarget.getAlias()).append("]\n");
        articles.forEach(article -> sb.append(article.convertArticletoMessage()).append("\n"));
        sb.append("-----------------------------------\n");

        ChatPostMessageRequest request = userNotifyService.generateMessage(sb.toString()).build();
        return userNotifyService.notify(request);
    }

    private void notifyMessage(UserNotifyModel userNotifyModel) {
        System.out.println("article = " + userNotifyModel);
        if (userNotifyModel.getArticles().isEmpty()) {
            return;
        }

        ChatPostMessageRequest request = userNotifyService.generateMessage(userNotifyModel.toCompositedMessage())
                                                          .threadTs(userNotifyModel.getChatPostMessageResponse().getTs())
                                                          .build();
        userNotifyService.notify(request);
    }

}
